#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Data {
  [Item("StringArray", "Represents an array of strings.")]
  [StorableClass]
  public class StringArray : Item, IEnumerable<string>, IStringConvertibleArray {
    public static new Image StaticItemImage {
      get { return HeuristicLab.Common.Resources.VSImageLibrary.Class; }
    }

    [Storable]
    protected string[] array;

    [Storable]
    protected List<string> columnNames;
    public virtual IEnumerable<string> ColumnNames {
      get { return this.columnNames; }
      set {
        if (ReadOnly) throw new NotSupportedException("ColumnNames cannot be set. ValueTypeArray is read-only.");
        if (value == null || value.Count() == 0)
          columnNames = new List<string>();
        else if (value.Count() != Length)
          throw new ArgumentException("A column name must be specified for each column.");
        else
          columnNames = new List<string>(value);
        OnColumnNamesChanged();
      }
    }

    public virtual int Length {
      get { return array.Length; }
      protected set {
        if (ReadOnly) throw new NotSupportedException("Length cannot be set. StringArray is read-only.");
        if (value != Length) {
          Array.Resize<string>(ref array, value);
          while (columnNames.Count > value)
            columnNames.RemoveAt(columnNames.Count - 1);
          while (columnNames.Count < value)
            columnNames.Add("Column " + columnNames.Count);
          OnColumnNamesChanged();
          OnReset();
        }
      }
    }
    public virtual string this[int index] {
      get { return array[index]; }
      set {
        if (ReadOnly) throw new NotSupportedException("Item cannot be set. StringArray is read-only.");
        if (value != array[index]) {
          if ((value != null) || (array[index] != string.Empty)) {
            array[index] = value != null ? value : string.Empty;
            OnItemsChanged(new List<int>(1) { index });
          }
        }
      }
    }

    [Storable]
    protected bool readOnly;
    public virtual bool ReadOnly {
      get { return readOnly; }
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      if (columnNames == null) {
        columnNames = new List<string>(array.Length);
        while (columnNames.Count < array.Length)
          columnNames.Add("Column " + columnNames.Count);
      }
    }

    [StorableConstructor]
    protected StringArray(bool deserializing) : base(deserializing) { }
    protected StringArray(StringArray original, Cloner cloner)
      : base(original, cloner) {
      this.array = (string[])original.array.Clone();
      this.columnNames = new List<string>(original.columnNames);
      this.readOnly = original.readOnly;
    }
    public StringArray() {
      array = new string[0];
      columnNames = new List<string>();
      readOnly = false;
    }
    public StringArray(int length) {
      array = new string[length];
      for (int i = 0; i < array.Length; i++)
        array[i] = string.Empty;
      columnNames = new List<string>();
      readOnly = false;
    }
    public StringArray(string[] elements) {
      if (elements == null) throw new ArgumentNullException();
      array = new string[elements.Length];
      for (int i = 0; i < array.Length; i++)
        array[i] = elements[i] == null ? string.Empty : elements[i];
      columnNames = new List<string>();
      readOnly = false;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new StringArray(this, cloner);
    }

    public virtual StringArray AsReadOnly() {
      StringArray readOnlyStringArray = (StringArray)this.Clone();
      readOnlyStringArray.readOnly = true;
      return readOnlyStringArray;
    }

    public override string ToString() {
      StringBuilder sb = new StringBuilder();
      sb.Append("[");
      if (array.Length > 0) {
        sb.Append(array[0]);
        for (int i = 1; i < array.Length; i++)
          sb.Append(";").Append(array[i]);
      }
      sb.Append("]");
      return sb.ToString();
    }

    public virtual IEnumerator<string> GetEnumerator() {
      return array.Cast<string>().GetEnumerator();
    }

    IEnumerator IEnumerable.GetEnumerator() {
      return GetEnumerator();
    }

    protected virtual bool Validate(string value, out string errorMessage) {
      if (value == null) {
        errorMessage = "Invalid Value (string must not be null)";
        return false;
      } else {
        errorMessage = string.Empty;
        return true;
      }
    }
    protected virtual string GetValue(int index) {
      return this[index];
    }
    protected virtual bool SetValue(string value, int index) {
      if (ReadOnly) throw new NotSupportedException("Item cannot be set. StringArray is read-only.");
      if (value != null) {
        this[index] = value;
        return true;
      } else {
        return false;
      }
    }
    public virtual bool SetValues(IEnumerable<ArrayValue<string>> arrayValues) {
      if (ReadOnly) throw new NotSupportedException("Item cannot be set. StringArray is read-only.");
      if (!arrayValues.All(v => v.Value != null)) return false;
      List<int> positions = new List<int>();
      foreach (var v in arrayValues) {
        if (array[v.Index] != v.Value) {
          array[v.Index] = v.Value;
          positions.Add(v.Index);
        }
      }
      OnItemsChanged(positions);
      return true;
    }

    public event EventHandler ColumnNamesChanged;
    protected virtual void OnColumnNamesChanged() {
      EventHandler handler = ColumnNamesChanged;
      if (handler != null)
        handler(this, EventArgs.Empty);
    }
    public event EventHandler<ArrayValuesChangedEventArgs> ItemsChanged;
    protected virtual void OnItemsChanged(IEnumerable<int> indices) {
      if (ItemsChanged != null)
        ItemsChanged(this, new ArrayValuesChangedEventArgs(indices));
      OnToStringChanged();
    }
    public event EventHandler Reset;
    protected virtual void OnReset() {
      if (Reset != null)
        Reset(this, EventArgs.Empty);
      OnToStringChanged();
    }

    #region IStringConvertibleArray Members
    int IStringConvertibleArray.Length {
      get { return Length; }
      set { Length = value; }
    }
    bool IStringConvertibleArray.Validate(string value, out string errorMessage) {
      return Validate(value, out errorMessage);
    }
    string IStringConvertibleArray.GetValue(int index) {
      return GetValue(index);
    }
    bool IStringConvertibleArray.SetValue(string value, int index) {
      return SetValue(value, index);
    }
    bool IStringConvertibleArray.SetValue(ArrayValue<string> arrayValue) {
      return SetValue(arrayValue.Value, arrayValue.Index);
    }
    bool IStringConvertibleArray.SetValues(IEnumerable<ArrayValue<string>> arrayValues) {
      return SetValues(arrayValues);
    }
    #endregion
  }
}
