#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Text;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Data {
  [Item("ValueTypeMatrix", "An abstract base class for representing matrices of value types.")]
  [StorableClass]
  public abstract class ValueTypeMatrix<T> : Item, IEnumerable<T>, IStringConvertibleMatrix where T : struct {
    public static new Image StaticItemImage {
      get { return HeuristicLab.Common.Resources.VSImageLibrary.Class; }
    }

    [Storable]
    protected T[,] matrix;

    [Storable]
    protected List<string> columnNames;
    public virtual IEnumerable<string> ColumnNames {
      get { return this.columnNames; }
      set {
        if (ReadOnly) throw new NotSupportedException("ColumnNames cannot be set. ValueTypeMatrix is read-only.");
        if (value == null || value.Count() == 0)
          columnNames = new List<string>();
        else if (value.Count() != Columns)
          throw new ArgumentException("A column name must be specified for each column.");
        else
          columnNames = new List<string>(value);
        OnColumnNamesChanged();
      }
    }
    [Storable]
    protected List<string> rowNames;
    public virtual IEnumerable<string> RowNames {
      get { return this.rowNames; }
      set {
        if (ReadOnly) throw new NotSupportedException("RowNames cannot be set. ValueTypeMatrix is read-only.");
        if (value == null || value.Count() == 0)
          rowNames = new List<string>();
        else if (value.Count() != Rows)
          throw new ArgumentException("A row name must be specified for each row.");
        else
          rowNames = new List<string>(value);
        OnRowNamesChanged();
      }
    }
    [Storable]
    protected bool sortableView;
    public virtual bool SortableView {
      get { return sortableView; }
      set {
        if (ReadOnly) throw new NotSupportedException("SortableView cannot be set. ValueTypeMatrix is read-only.");
        if (value != sortableView) {
          sortableView = value;
          OnSortableViewChanged();
        }
      }
    }

    public virtual int Rows {
      get { return matrix.GetLength(0); }
      protected set {
        if (ReadOnly) throw new NotSupportedException("Rows cannot be set. ValueTypeMatrix is read-only.");
        if (value != Rows) {
          T[,] newArray = new T[value, Columns];
          Array.Copy(matrix, newArray, Math.Min(value * Columns, matrix.Length));
          matrix = newArray;
          while (rowNames.Count > value)
            rowNames.RemoveAt(rowNames.Count - 1);
          while (rowNames.Count < value)
            rowNames.Add("Row " + rowNames.Count);
          OnRowsChanged();
          OnRowNamesChanged();
          OnReset();
        }
      }
    }
    public virtual int Columns {
      get { return matrix.GetLength(1); }
      protected set {
        if (ReadOnly) throw new NotSupportedException("Columns cannot be set. ValueTypeMatrix is read-only.");
        if (value != Columns) {
          T[,] newArray = new T[Rows, value];
          for (int i = 0; i < Rows; i++)
            Array.Copy(matrix, i * Columns, newArray, i * value, Math.Min(value, Columns));
          matrix = newArray;
          while (columnNames.Count > value)
            columnNames.RemoveAt(columnNames.Count - 1);
          while (columnNames.Count < value)
            columnNames.Add("Column " + columnNames.Count);
          OnColumnsChanged();
          OnColumnNamesChanged();
          OnReset();
        }
      }
    }
    public virtual T this[int rowIndex, int columnIndex] {
      get { return matrix[rowIndex, columnIndex]; }
      set {
        if (ReadOnly) throw new NotSupportedException("Item cannot be set. ValueTypeMatrix is read-only.");
        if (!value.Equals(matrix[rowIndex, columnIndex])) {
          matrix[rowIndex, columnIndex] = value;
          OnItemsChanged(new List<MatrixPosition>(1) { new MatrixPosition(rowIndex, columnIndex) });
        }
      }
    }

    public virtual void SetValues(IEnumerable<MatrixValue<T>> matrixValues) {
      if (ReadOnly) throw new NotSupportedException("Item cannot be set. ValueTypeMatrix is read-only.");
      List<MatrixPosition> positions = new List<MatrixPosition>();
      foreach (var matrixValue in matrixValues) {
        if (!matrixValue.Value.Equals(matrix[matrixValue.Position.Row, matrixValue.Position.Column])) {
          matrix[matrixValue.Position.Row, matrixValue.Position.Column] = matrixValue.Value;
          positions.Add(matrixValue.Position);
        }
      }
      OnItemsChanged(positions);
    }

    [Storable]
    protected bool readOnly;
    public virtual bool ReadOnly {
      get { return readOnly; }
    }

    [StorableConstructor]
    protected ValueTypeMatrix(bool deserializing) : base(deserializing) { }
    protected ValueTypeMatrix(ValueTypeMatrix<T> original, Cloner cloner)
      : base(original, cloner) {
      this.matrix = (T[,])original.matrix.Clone();
      this.columnNames = new List<string>(original.columnNames);
      this.rowNames = new List<string>(original.rowNames);
      this.sortableView = original.sortableView;
      this.readOnly = original.readOnly;
    }
    protected ValueTypeMatrix() {
      matrix = new T[0, 0];
      columnNames = new List<string>();
      rowNames = new List<string>();
      sortableView = false;
      readOnly = false;
    }
    protected ValueTypeMatrix(int rows, int columns) {
      matrix = new T[rows, columns];
      columnNames = new List<string>();
      rowNames = new List<string>();
      sortableView = false;
      readOnly = false;
    }
    protected ValueTypeMatrix(int rows, int columns, IEnumerable<string> columnNames)
      : this(rows, columns) {
      ColumnNames = columnNames;
    }
    protected ValueTypeMatrix(int rows, int columns, IEnumerable<string> columnNames, IEnumerable<string> rowNames)
      : this(rows, columns, columnNames) {
      RowNames = rowNames;
    }
    protected ValueTypeMatrix(T[,] elements) {
      if (elements == null) throw new ArgumentNullException();
      matrix = (T[,])elements.Clone();
      columnNames = new List<string>();
      rowNames = new List<string>();
      sortableView = false;
      readOnly = false;
    }
    protected ValueTypeMatrix(T[,] elements, IEnumerable<string> columnNames)
      : this(elements) {
      ColumnNames = columnNames;
    }
    protected ValueTypeMatrix(T[,] elements, IEnumerable<string> columnNames, IEnumerable<string> rowNames)
      : this(elements, columnNames) {
      RowNames = rowNames;
    }

    public virtual ValueTypeMatrix<T> AsReadOnly() {
      ValueTypeMatrix<T> readOnlyValueTypeMatrix = (ValueTypeMatrix<T>)this.Clone();
      readOnlyValueTypeMatrix.readOnly = true;
      return readOnlyValueTypeMatrix;
    }

    public override string ToString() {
      StringBuilder sb = new StringBuilder();
      sb.Append("[");
      if (matrix.Length > 0) {
        for (int i = 0; i < Rows; i++) {
          sb.Append("[").Append(matrix[i, 0].ToString());
          for (int j = 1; j < Columns; j++)
            sb.Append(";").Append(matrix[i, j].ToString());
          sb.Append("]");
        }
      }
      sb.Append("]");
      return sb.ToString();
    }

    public virtual IEnumerator<T> GetEnumerator() {
      return matrix.Cast<T>().GetEnumerator();
    }

    IEnumerator IEnumerable.GetEnumerator() {
      return GetEnumerator();
    }

    #region events
    public event EventHandler ColumnsChanged;
    protected virtual void OnColumnsChanged() {
      EventHandler handler = ColumnsChanged;
      if (handler != null)
        handler(this, EventArgs.Empty);
    }
    public event EventHandler RowsChanged;
    protected virtual void OnRowsChanged() {
      EventHandler handler = RowsChanged;
      if (handler != null)
        handler(this, EventArgs.Empty);
    }
    public event EventHandler ColumnNamesChanged;
    protected virtual void OnColumnNamesChanged() {
      EventHandler handler = ColumnNamesChanged;
      if (handler != null)
        handler(this, EventArgs.Empty);
    }
    public event EventHandler RowNamesChanged;
    protected virtual void OnRowNamesChanged() {
      EventHandler handler = RowNamesChanged;
      if (handler != null)
        handler(this, EventArgs.Empty);
    }
    public event EventHandler SortableViewChanged;
    protected virtual void OnSortableViewChanged() {
      EventHandler handler = SortableViewChanged;
      if (handler != null)
        handler(this, EventArgs.Empty);
    }
    public event EventHandler<MatrixValuesChangedEventArgs> ItemsChanged;
    protected virtual void OnItemsChanged(IEnumerable<MatrixPosition> positions) {
      if (ItemsChanged != null)
        ItemsChanged(this, new MatrixValuesChangedEventArgs(positions));
      OnToStringChanged();
    }
    public event EventHandler Reset;
    protected virtual void OnReset() {
      if (Reset != null)
        Reset(this, EventArgs.Empty);
      OnToStringChanged();
    }
    #endregion

    protected abstract bool Validate(string value, out string errorMessage);
    protected abstract bool SetValue(string value, int rowIndex, int columnIndex);
    protected abstract string GetValue(int rowIndex, int columIndex);
    protected abstract bool SetValues(IEnumerable<MatrixValue<string>> rowColumnValues);

    #region IStringConvertibleMatrix Members
    int IStringConvertibleMatrix.Rows {
      get { return Rows; }
      set { Rows = value; }
    }
    int IStringConvertibleMatrix.Columns {
      get { return Columns; }
      set { Columns = value; }
    }

    bool IStringConvertibleMatrix.Validate(string value, out string errorMessage) {
      return Validate(value, out errorMessage);
    }
    string IStringConvertibleMatrix.GetValue(int rowIndex, int columnIndex) {
      return GetValue(rowIndex, columnIndex);
    }
    bool IStringConvertibleMatrix.SetValue(string value, int rowIndex, int columnIndex) {
      return SetValue(value, rowIndex, columnIndex);
    }

    bool IStringConvertibleMatrix.SetValue(MatrixValue<string> rowColumnValue) {
      return SetValue(rowColumnValue.Value, rowColumnValue.Position.Row, rowColumnValue.Position.Column);
    }

    bool IStringConvertibleMatrix.SetValues(IEnumerable<MatrixValue<string>> rowColumnValues) {
      return SetValues(rowColumnValues);
    }
    #endregion
  }
}
